//wqpnu: complex rational class etc.
 
#include <vcl.h>
#include <iostream.h>


namespace {
    struct Init_DecimalSeparator
    {
        Init_DecimalSeparator(){DecimalSeparator='.';}
    };
    Init_DecimalSeparator Init_DecimalSeparator_dummy; //zum Aufruf des Konstruktors, welcher initialisiert
}

//local defs, so nu.. prefix:
const int nu_max_int = std::numeric_limits < int >::max() - 1;
const int nu_min_int = std::numeric_limits < int >::min() + 1;


void floerr(); //e.g. for testing catch

//ggt=groesster gemeinsamer Teiler
int ggt(int a, int b);

//rational class: ratyp        ***************************
class ratyp
{
    int nu; //Numerator
    int de; //Denominator
    
    bool isra; //true if nu/de valid, false if qo valid
    
    double qo; //approximation of nu/de in case of invalid nu/de
    
    void shorten();
    public:
    ratyp() : nu(0), de(1), isra(true) {}
    ratyp(int n) : nu(n), de(1), isra(true) {}
    ratyp(int n, int d) : nu(n), de(d), isra(true) { shorten(); }
    
    /*
    //3 alternative constructors only for test of qo evaluation
    ratyp() : qo(1), isra(false) {}
    ratyp(int n) : isra(false) {qo=double(n);}
    ratyp(int n, int d) : isra(false) { qo=double(n)/double(d); }
    */
    
    //in case of overflow or float input approximation by double:
    ratyp(double q) : qo(q), isra(false) {}
    ratyp& operator = (int n) { return assign(n); }
    ratyp& operator = (double x) { return assign(x); }
    
    // Assign in place
    ratyp& assign(int n, int d);
    ratyp& assign(int n);
    ratyp& assign(double x);
    
    // Access to representation
    int num() const { return nu; }
    int den() const { return de; }
    bool israt() const { return isra; }
    double quot() const { return qo; }
    double florep() const
    { return isra ? double(nu)/double(de) : qo; }
    
    bool InputFromStr(const AnsiString& s); //true, if input from AnsiString ok
    
    // Arithmetic assignment operators
    ratyp& operator += (const ratyp& r);
    ratyp& operator -= (const ratyp& r);
    ratyp& operator*= (const ratyp& r);
    ratyp& operator/= (const ratyp& r);
    
    //mixed mode:
    ratyp& operator += (int i){return operator += (ratyp(i)); }
    ratyp& operator -= (int i){return operator -= (ratyp(i)); }
    ratyp& operator*= (int i){return operator*= (ratyp(i)); }
    ratyp& operator/= (int i){return operator/= (ratyp(i)); }
    ratyp& operator += (double x){return operator += (ratyp(x)); }
    ratyp& operator -= (double x){return operator -= (ratyp(x)); }
    ratyp& operator*= (double x){return operator*= (ratyp(x)); }
    ratyp& operator/= (double x){return operator/= (ratyp(x)); }
    
    // Increment and decrement
    const ratyp& operator++();
    const ratyp& operator--();
    const ratyp operator++(int);
    const ratyp operator--(int);
    
    // Comparison operators
    bool operator < (const ratyp& r) const;
    bool operator < (const double& x) const;
    bool operator < (int i) const;
    bool operator == (const ratyp& r) const;
    bool operator == (const double& x) const;
    bool operator == (int i) const;
    bool operator > (const ratyp& r) const;
    bool operator > (const double& x) const;
    bool operator > (int i) const;
    
    bool operator !() const { return isra ? !nu : !qo; };
    
    operator AnsiString() const;
};

inline ratyp& ratyp::assign(int n, int d)
{
    nu = n;
    de = d;
    shorten();
    isra = true;
    return *this;
}

inline ratyp& ratyp::assign(int n)
{
    nu = n;
    de = 1;
    isra = true;
    return *this;
}

inline ratyp& ratyp::assign(double x)
{
    qo = x;
    isra = false;
    return *this;
}

bool ratyp::operator == (const double& x) const
{
    if (isra)
        return (double(nu)/double(de) == x);
    else
        return qo == x;
}

inline bool ratyp::operator == (int i) const
{
    if (isra)
        return ((de == int(1)) && (nu == i));
    else
        return qo == double(i);
}


//implicit conversion would be in class ratyp:
//    operator double() const {return isra?double(nu)/double(de):qo;}
//... but explicit conversion is better to avoid ambiguity:
//inline double flotyp(const ratyp& r) {return r.israt() ? double(r.num())/double(r.den()):r.quot(); }

// Unary plus and minus

inline ratyp operator + (const ratyp& r)
{ return r; }

inline ratyp operator - (const ratyp& r)
{
    if (r.israt())
        return ratyp(- r.num(), r.den());
    else
        return ratyp(- r.quot());
}
//mixed mode; preferences of templates less than of explicit functions
//cf Stoustrup 2000, S. 359
template < typename T >
inline bool operator >= (const T& x, const ratyp& y)  {return !(y > x); }
template < typename T >
inline bool operator <= (const T& x, const ratyp& y) { return !(y < x); }
template < typename T >
inline bool operator >(const T& x, const ratyp& y) { return (y < x); }
template < typename T >
inline bool operator < (const T& x, const ratyp& y) { return (y > x); }

//Utilities
int kgv(int n, int m);
ratyp abs(const ratyp& r);

//class cotyp; complex type which uses ratyp ***************************
class cotyp
{
    ratyp re; //real part
    ratyp im; //imaginary part
    public:
    cotyp() : re(0), im(0) {}
    cotyp(ratyp x) : re(x), im(0) {}
    
    
    //distinctions ratyp - cotyp to avoid ambiguities when using numbers in code
    //if cotyp_as_default defined, default number format is cotyp
    //#define cotyp_as_default
    
    #ifdef cotyp_as_default
    cotyp(double x) : re(x), im(0) {}
    cotyp(int x) : re(x), im(0) {}
    #endif
    cotyp(ratyp re_, ratyp im_) : re(re_), im(im_) {}
    cotyp(int renu, int rede, int imnu, int imde) :
    re(renu, rede), im(imnu, imde) {}
    
    cotyp& operator = (ratyp n) { return assign(n); }
    //    cotyp& operator=(double x) { return assign(x); }
    //    cotyp& operator=(int i) { return assign(i); }
    
    // Assign in place
    cotyp& assign(int renu, int rede, int imnu, int imde)
    {re.assign(renu, rede); im.assign(imnu, imde); return *this; }
    cotyp& assign(ratyp r, ratyp i) { re = r; im = i; return *this; }
    cotyp& assign(ratyp r) { re = r; im = 0; return *this; }
    //cotyp& assign(double r) { re = r; im = 0; return *this; };
    
    // Access to representation
    ratyp real() const { return re; }
    ratyp imag() const { return im; }
    
    cotyp florep() const
    {cotyp c; c.re = re.florep();
        c.im = im.florep(); return c; }
    
    bool InputFromStr(const AnsiString& s); //true, if input from AnsiString ok
    
    // Arithmetic assignment operators
    cotyp& operator += (const cotyp& c);
    cotyp& operator -= (const cotyp& c);
    cotyp& operator*= (const cotyp& c);
    cotyp& operator/= (const cotyp& c);
    
    //mixed mode:
    cotyp& operator += (int i){return operator += (cotyp(i)); }
    cotyp& operator -= (int i){return operator -= (cotyp(i)); }
    cotyp& operator*= (int i){return operator*= (cotyp(i)); }
    cotyp& operator/= (int i){return operator/= (cotyp(i)); }
    cotyp& operator += (ratyp r){return operator += (cotyp(r)); }
    cotyp& operator -= (ratyp r){return operator -= (cotyp(r)); }
    cotyp& operator*= (ratyp r){return operator*= (cotyp(r)); }
    cotyp& operator/= (ratyp r){return operator/= (cotyp(r)); }
    cotyp& operator += (double x){return operator += (cotyp(x)); }
    cotyp& operator -= (double x){return operator -= (cotyp(x)); }
    cotyp& operator*= (double x){return operator*= (cotyp(x)); }
    cotyp& operator/= (double x){return operator/= (cotyp(x)); }
    
    const cotyp& cotyp::operator++() {re++; return *this; }
    const cotyp& cotyp::operator--() {re--; return *this; }
    const cotyp cotyp::operator++(int)
    { cotyp tmp(*this); ++(*this); return tmp; }
    const cotyp cotyp::operator--(int)
    { cotyp tmp(*this); --(*this); return tmp; }
    
    bool cotyp::operator == (const cotyp& c) const
    {return re == c.re && im == c.im; }
    bool cotyp::operator == (const ratyp& x) const
    {return re == x && im == 0; }
    operator AnsiString() const;
};

// Unary plus and minus

inline cotyp operator + (const cotyp& c)
{ return c; }

inline cotyp operator - (const cotyp& c)
{ return cotyp(- c.real(), - c.imag()); }

ratyp abs2(const cotyp& c);

AnsiString astr(int letters, const ratyp& r);//length of astr dependent on letters
AnsiString astr(int letters, const cotyp& c);


#ifdef cotyp_as_default

template < typename T >
inline bool operator != (const T& x, const cotyp& y) { return !(y == x); }
template < typename T >
inline bool operator == (const T& x, const cotyp& y) { return (y == x); }
template < typename A >
inline cotyp operator + (const A& a, const cotyp& b)
{cotyp aa = a; return cotyp(aa) += b; }
template < typename A >
inline cotyp operator - (const A& a, const cotyp& b)
{cotyp aa = a; return cotyp(aa) -= b; }
template < typename A >
inline cotyp operator* (const A& a, const cotyp& b)
{cotyp aa = a; return cotyp(aa) *= b; }
template < typename A >
inline cotyp operator/ (const A& a, const cotyp& b)
{cotyp aa = a; return cotyp(aa) /= b; }

#else
    
    template < typename T >
    inline bool operator != (const T& x, const ratyp& y) { return !(y == x); }
template < typename T >
inline bool operator == (const T& x, const ratyp& y) { return (y == x); }
template < typename A >
inline ratyp operator + (const A& a, const ratyp& b)
{ratyp aa = a; return ratyp(aa) += b; }
template < typename A >
inline ratyp operator - (const A& a, const ratyp& b)
{ratyp aa = a; return ratyp(aa) -= b; }
template < typename A >
inline ratyp operator* (const A& a, const ratyp& b)
{ratyp aa = a; return ratyp(aa) *= b; }
template < typename A >
inline ratyp operator/ (const A& a, const ratyp& b)
{ratyp aa = a; return ratyp(aa) /= b; }

#endif






